#define INCL_DOS
#define INCL_WIN
#define INCL_GPI

#include <os2.h>
#include "dw.h"

#define  _MEERROR_H_
#include <mmioos2.h>
#include <dive.h>
#include <fourcc.h>

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "format.h"
#include "decoder_plug.h"
#include "plugin.h"

#define FULL_CLEAR

HWND hwndClient, hwndPM123;

#define strcasecmp stricmp

PLUGIN_PROCS procs;
int (* _System specana_init)(int setnumsamples);
int (* _System specana_dobands)(float bands[]);

void DiveInit(void);
void DiveShutdown(void);

HAB hab;

/* So I can save/load this information */
int blockwidth = 2;
int falloff = 20;
int updatefreq = 32;
int spacing = 1;
int percentage = 20;
int boostval = 3;
int fire = 2;
int usedive = TRUE;
int showpeaks = TRUE;
int gradient = TRUE;
int boost = TRUE;
int barred = FALSE;
int barblue = FALSE;
int bargreen = TRUE;

int bands, samples;
int cx, cy;

int *maxage;
float *bmax;
float *last;
float *current;

DIVE_CAPS DiveCaps = {0};
FOURCC fccFormats[100] = {0};
HDIVE hDive = NULLHANDLE;
PBYTE fBuf;
ULONG DiveBuffer = 0, fBufLen = 0, fBufScanLineBytes = 0, fBufScanLines = 0;
RGB2 *DiveColors = NULL;
RGB2 rgbBlack = { 0, 0, 0, 0};
RGB2 rgbWhite = { 255, 255, 255, 0};
HPAL hpalDive;
HDC hdcClient;
HPS hpsDive;

int _System plugin_query(PPLUGIN_QUERYPARAM param)
{
	param->type         = PLUGIN_VISUAL; /* Identify the plugin as visual */

	param->author       = "Brian Smith";
	/* Author of the plugin */

	param->desc         = "Nuclear Visuals";
	/* A short description of the plugin */

	param->configurable = TRUE;
	/* Toggles plugin configurability via PM123 Properties dialog */

	return 0;
}

/* Write the nuclear.ini file with all of the current settings */
void saveconfig(void)
{
	FILE *f;

	if((f=fopen("nuclear.ini", "w"))==NULL)
	{
		dw_messagebox("Nuclear Visuals", DW_MB_OK | DW_MB_ERROR, "Could not save settings.");
		return;
	}

	if(showpeaks == TRUE)
		fprintf(f, "SHOWPEAKS=TRUE\n");
	else
		fprintf(f, "SHOWPEAKS=FALSE\n");

	if(gradient == TRUE)
		fprintf(f, "GRADIENT=TRUE\n");
	else
		fprintf(f, "GRADIENT=FALSE\n");

	if(boost == TRUE)
		fprintf(f, "BOOST=TRUE\n");
	else
		fprintf(f, "BOOST=FALSE\n");

	if(usedive == TRUE)
		fprintf(f, "USEDIVE=TRUE\n");
	else
		fprintf(f, "USEDIVE=FALSE\n");

	if(barred == TRUE)
		fprintf(f, "BARRED=TRUE\n");
	else
		fprintf(f, "BARRED=FALSE\n");

	if(bargreen == TRUE)
		fprintf(f, "BARGREEN=TRUE\n");
	else
		fprintf(f, "BARGREEN=FALSE\n");

	if(barblue == TRUE)
		fprintf(f, "BARBLUE=TRUE\n");
	else
		fprintf(f, "BARBLUE=FALSE\n");

	fprintf(f, "FALLOFF=%d\n", falloff);
	fprintf(f, "UPDATEFREQ=%d\n", updatefreq);
	fprintf(f, "BLOCKWIDTH=%d\n", blockwidth);
	fprintf(f, "SPACING=%d\n", spacing);
	fprintf(f, "PERCENTAGE=%d\n", percentage);
	fprintf(f, "BOOSTVAL=%d\n", boostval);
	fprintf(f, "FIRE=%d\n", fire);

	fclose(f);
}

/* Generic function to parse information from a config file */
void getline(FILE *f, char *entry, char *entrydata)
{
	char in[256];
	int z;

	memset(in, 0, 256);
	fgets(in, 255, f);

	if(in[strlen(in)-1] == '\n')
		in[strlen(in)-1] = 0;

	if(in[0] != '#')
	{
		for(z=0;z<strlen(in);z++)
		{
			if(in[z] == '=')
			{
				in[z] = 0;
				strcpy(entry, in);
				strcpy(entrydata, &in[z+1]);
				return;
			}
		}
	}
	strcpy(entry, "");
	strcpy(entrydata, "");
}

/* Load the nuclear.ini file from disk setting all the necessary flags */
void loadconfig(void)
{
	char entry[256], entrydata[256];
	FILE *f;

	if((f = fopen("nuclear.ini", "r"))==NULL)
		return;

	while(!feof(f))
	{
		getline(f, entry, entrydata);
		if(strcasecmp(entry, "showpeaks")==0 && strcasecmp(entrydata, "true") == 0)
			showpeaks = TRUE;
		if(strcasecmp(entry, "gradient")==0 && strcasecmp(entrydata, "true") == 0)
			gradient = TRUE;
		if(strcasecmp(entry, "boost")==0 && strcasecmp(entrydata, "true") == 0)
			boost = TRUE;
		if(strcasecmp(entry, "usedive")==0 && strcasecmp(entrydata, "true") == 0)
			usedive = TRUE;
		if(strcasecmp(entry, "barred")==0 && strcasecmp(entrydata, "true") == 0)
			barred = TRUE;
		if(strcasecmp(entry, "bargreen")==0 && strcasecmp(entrydata, "true") == 0)
			bargreen = TRUE;
		if(strcasecmp(entry, "barblue")==0 && strcasecmp(entrydata, "true") == 0)
			barblue = TRUE;
		if(strcasecmp(entry, "showpeaks")==0 && strcasecmp(entrydata, "false") == 0)
			showpeaks = FALSE;
		if(strcasecmp(entry, "gradient")==0 && strcasecmp(entrydata, "false") == 0)
			gradient = FALSE;
		if(strcasecmp(entry, "boost")==0 && strcasecmp(entrydata, "false") == 0)
			boost = FALSE;
		if(strcasecmp(entry, "usedive")==0 && strcasecmp(entrydata, "false") == 0)
			usedive = FALSE;
		if(strcasecmp(entry, "barred")==0 && strcasecmp(entrydata, "false") == 0)
			barred = FALSE;
		if(strcasecmp(entry, "bargreen")==0 && strcasecmp(entrydata, "false") == 0)
			bargreen = FALSE;
		if(strcasecmp(entry, "barblue")==0 && strcasecmp(entrydata, "false") == 0)
			barblue = FALSE;
		if(strcasecmp(entry, "falloff")==0)
			falloff = atoi(entrydata);
		if(strcasecmp(entry, "updatefreq")==0)
			updatefreq = atoi(entrydata);
		if(strcasecmp(entry, "blockwidth")==0)
			blockwidth = atoi(entrydata);
		if(strcasecmp(entry, "spacing")==0)
			spacing = atoi(entrydata);
		if(strcasecmp(entry, "percentage")==0)
			percentage = atoi(entrydata);
		if(strcasecmp(entry, "boostval")==0)
			boostval = atoi(entrydata);
		if(strcasecmp(entry, "fire")==0)
			fire = atoi(entrydata);
	}
	fclose(f);
}


void DWSIGNAL generic_cancel(HWND window, void *data)
{
	HWND *handles = (HWND *)data;

	if(handles)
	{
		dw_window_destroy(handles[0]);
		free(handles);
	}
}

/* Fills the analyzer area with black. */
void clear_analyzer(void)
{
	if(usedive)
	{
		DiveBeginImageBufferAccess(hDive, DiveBuffer, &fBuf, &fBufScanLineBytes, &fBufScanLines);
		fBufLen = fBufScanLineBytes * fBufScanLines;

		memset(fBuf, 0, fBufLen);
		DiveBlitImage(hDive,
					  DiveBuffer,
					  DIVE_BUFFER_SCREEN);
		DiveEndImageBufferAccess(hDive, DiveBuffer);
	}
	else
	{
		HPS hps = WinGetPS(hwndClient);
		POINTL ptl;

		GpiSetColor(hps, CLR_BLACK);

		ptl.x = 0;
		ptl.y = 0;

		GpiMove(hps, &ptl);

		ptl.x = cx;
		ptl.y = cy;

		GpiBox(hps, DRO_FILL, &ptl, 0, 0);

		WinReleasePS(hps);
	}
}

/* Stops the analyzers, updates all the settings
 * and then restarts it.
 */
void update_analyzer(HWND *handles)
{
	int z = 1, wantedsamples;

	bands = 0;

	WinStopTimer(hab, hwndClient, 449);

	if(usedive)
		DiveShutdown();

	showpeaks = dw_checkbox_query(handles[1]);
	gradient = dw_checkbox_query(handles[2]);
	boost = dw_checkbox_query(handles[11]);
	usedive = dw_checkbox_query(handles[13]);
	barred = dw_checkbox_query(handles[3]);
	bargreen = dw_checkbox_query(handles[4]);
	barblue = dw_checkbox_query(handles[5]);
	falloff = dw_spinbutton_query(handles[6]);
	updatefreq = dw_spinbutton_query(handles[7]);
	blockwidth = dw_spinbutton_query(handles[8]);
	spacing = dw_spinbutton_query(handles[9]);
	percentage = dw_spinbutton_query(handles[10]);
	boostval = dw_spinbutton_query(handles[12]);

	if(dw_checkbox_query(handles[15]))
		fire = 1;
	else if(dw_checkbox_query(handles[14]))
		fire = 2;
	else
		fire = 0;

	saveconfig();

	/* Recalculate the sizes again */
	wantedsamples = (int)((cx / (blockwidth+spacing))*(100/percentage));
	samples = (int)(cx / (blockwidth+spacing));

	/* Find the ideal number of bars (bands) for our analyzer */
	while(bands < wantedsamples)
	{
		bands = specana_init(z);
		z++;
		if(z < 1 || z > 4097)
		{
			dw_messagebox("Nuclear Visuals", DW_MB_OK | DW_MB_ERROR, "Error getting spectrum analyzer");
			samples = 0;
			break;
		}
	}

    /* Free the old arrays */
	free(last);
	free(bmax);
	free(current);
	free(maxage);
	if(DiveColors)
	{
		free(DiveColors);
		DiveColors = NULL;
	}

	/* Reallocate them with the new size */
	maxage = (int *)calloc(bands, sizeof(int));
	bmax = (float *)calloc(bands, sizeof(float));
	last = (float *)calloc(bands, sizeof(float));
	current = (float *)calloc(bands, sizeof(float));
	if(usedive)
		DiveColors = (RGB2 *)calloc(cy+2, sizeof(RGB2));

	if(usedive)
		DiveInit();

	clear_analyzer();

	WinStartTimer(hab, hwndClient, 449, updatefreq);
}

/* Callback to handle an OK press in the preferences dialog */
void DWSIGNAL preferences_ok(HWND window, void *data)
{
	HWND *handles = (HWND *)data;

	if(handles)
	{
		update_analyzer(handles);

		dw_window_destroy(handles[0]);
		free(handles);
	}
}

/* Callback to handle an Apply press in the preferences dialog */
void DWSIGNAL preferences_apply(HWND window, void *data)
{
	HWND *handles = (HWND *)data;

	if(handles)
		update_analyzer(handles);
}

int _System plugin_configure(HWND hwnd, HMODULE module)
{
	HWND entrywindow, mainbox, vbox, cancelbutton, okbutton, buttonbox,
	applybutton, xbox, stext, winbox, groupbox, notebook,
	*handles = malloc(17 * sizeof(HWND));
	ULONG flStyle = DW_FCF_TITLEBAR | DW_FCF_SHELLPOSITION | DW_FCF_DLGBORDER;
	int pageid, tmp;

	handles[0] = entrywindow = dw_window_new(HWND_DESKTOP, "Nuclear Visuals Settings", flStyle);

	winbox = dw_box_new(BOXVERT, 0);

	dw_box_pack_start(entrywindow, winbox, 0, 0, TRUE, TRUE, 0);

	notebook = dw_notebook_new(1050L, TRUE);

	dw_box_pack_start(winbox, notebook, 150,300, TRUE, TRUE, 0);

	xbox = dw_box_new(BOXVERT, 0);

	mainbox = dw_box_new(BOXHORZ, 10);

	dw_box_pack_start(xbox, mainbox, 0, 0, TRUE, TRUE, 0);

	/* Checkboxes */
	vbox = dw_box_new(BOXVERT, 0);

	dw_box_pack_start(mainbox, vbox, 0, 0, TRUE, TRUE, 0);

	handles[1] = dw_checkbox_new("Show Peaks", 0);
	dw_checkbox_set(handles[1],showpeaks);


	dw_box_pack_start(vbox, handles[1], 50, 20, TRUE, TRUE, 4);

	handles[2] = dw_checkbox_new("Gradient", 0);
	dw_checkbox_set(handles[2],gradient);

	dw_box_pack_start(vbox, handles[2], 50, 20, TRUE, TRUE, 4);

	handles[3] = dw_checkbox_new("Bar Red", 0);
	dw_checkbox_set(handles[3],barred);

	dw_box_pack_start(vbox, handles[3], 50, 20, TRUE, TRUE, 4);

	handles[4] = dw_checkbox_new("Bar Green", 0);
	dw_checkbox_set(handles[4],bargreen);

	dw_box_pack_start(vbox, handles[4], 50, 20, TRUE, TRUE, 4);

	handles[5] = dw_checkbox_new("Bar Blue", 0);
	dw_checkbox_set(handles[5],barblue);

	dw_box_pack_start(vbox, handles[5], 50, 20, TRUE, TRUE, 4);

	handles[11] = dw_checkbox_new("Boost", 0);
	dw_checkbox_set(handles[11],boost);

	dw_box_pack_start(vbox, handles[11], 50, 20, TRUE, TRUE, 4);

	handles[13] = dw_checkbox_new("Use DIVE", 0);
	dw_checkbox_set(handles[13],usedive);

	dw_box_pack_start(vbox, handles[13], 50, 20, TRUE, TRUE, 4);

	/* Text */
	vbox = dw_box_new(BOXVERT, 0);

	dw_box_pack_start(mainbox, vbox, 0, 0, FALSE, TRUE, 0);

	stext = dw_text_new("Falloff", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	stext = dw_text_new("Update Frequency", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	stext = dw_text_new("Block Width", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	stext = dw_text_new("Spacing", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	stext = dw_text_new("Spectrum %", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	stext = dw_text_new("Boost Value", 0);

	dw_window_set_style(stext, DW_DT_VCENTER, DW_DT_VCENTER);

	dw_box_pack_start(vbox, stext, 100, 20, FALSE, TRUE, 4);

	dw_box_pack_start(vbox, 0, 100, 6, FALSE, TRUE, 0);

	/* Spinbuttons */
	vbox = dw_box_new(BOXVERT, 0);

	dw_box_pack_start(mainbox, vbox, 0, 0, TRUE, TRUE, 0);

	/* Falloff */
	handles[6] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[6], 60L, 0L);
	dw_spinbutton_set_pos(handles[6], falloff);

	dw_box_pack_start(vbox, handles[6], 25, 20, TRUE, TRUE, 8);

	/* Update Frequency */
	handles[7] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[7], 1000L, 1L);
	dw_spinbutton_set_pos(handles[7], updatefreq);

	dw_box_pack_start(vbox, handles[7], 25, 20, TRUE, TRUE, 8);

    /* Block Width */
	handles[8] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[8], 5L, 1L);
	dw_spinbutton_set_pos(handles[8], blockwidth);

	dw_box_pack_start(vbox, handles[8], 25, 20, TRUE, TRUE, 8);

    /* Spacing */
	handles[9] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[9], 5L, 0L);
	dw_spinbutton_set_pos(handles[9], spacing);

	dw_box_pack_start(vbox, handles[9], 25, 20, TRUE, TRUE, 8);

    /* Percentage */
	handles[10] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[10], 100L, 10L);
	dw_spinbutton_set_pos(handles[10], percentage);

	dw_box_pack_start(vbox, handles[10], 25, 20, TRUE, TRUE, 8);

    /* Boost Value */
	handles[12] = dw_spinbutton_new("", 100L);

	dw_spinbutton_set_limits(handles[12], 5L, 1L);
	dw_spinbutton_set_pos(handles[12], boostval);

	dw_box_pack_start(vbox, handles[12], 25, 20, TRUE, TRUE, 8);

	pageid = dw_notebook_page_new(notebook, 0L, FALSE);

	dw_notebook_pack(notebook, pageid, xbox);

	/* Due to a GTK limitiation the page text must be set after the page is packed */
	dw_notebook_page_set_text(notebook, pageid, "General");
	dw_notebook_page_set_status_text(notebook, pageid, "Page 1 of 2");

	xbox = dw_box_new(BOXVERT, 5);

	groupbox = dw_groupbox_new(BOXVERT, 10, "Fire Type");

	dw_box_pack_start(xbox, groupbox, 150, 70, TRUE, TRUE, 0);

	handles[14] = dw_radiobutton_new("Hot", 0);

	dw_box_pack_start(groupbox, handles[14], 50, 20, TRUE, TRUE, 4);

	handles[15] = dw_radiobutton_new("Cold", 0);

	dw_box_pack_start(groupbox, handles[15], 50, 20, TRUE, TRUE, 4);

	handles[16] = dw_radiobutton_new("Standard", 0);

	dw_box_pack_start(groupbox, handles[16], 50, 20, TRUE, TRUE, 4);

	pageid = dw_notebook_page_new(notebook, 0L, FALSE);

	dw_notebook_pack(notebook, pageid, xbox);

	/* Due to a GTK limitiation the page text must be set after the page is packed */
	dw_notebook_page_set_text(notebook, pageid, "Visuals");
	dw_notebook_page_set_status_text(notebook, pageid, "Page 2 of 2");

	/* Buttons */
	buttonbox = dw_box_new(BOXHORZ, 4);

	dw_box_pack_start(winbox, buttonbox, 0, 0, TRUE, TRUE, 4);

	okbutton = dw_button_new("Ok", 1001L);

	dw_box_pack_start(buttonbox, okbutton, 50, 40, TRUE, TRUE, 4);

	applybutton = dw_button_new("Apply", 1002L);

	dw_box_pack_start(buttonbox, applybutton, 50, 40, TRUE, TRUE, 4);

	cancelbutton = dw_button_new("Cancel", 1003L);

	dw_box_pack_start(buttonbox, cancelbutton, 50, 40, TRUE, TRUE, 4);

	dw_signal_connect(okbutton, "clicked", DW_SIGNAL_FUNC(preferences_ok), (void *)handles);
	dw_signal_connect(applybutton, "clicked", DW_SIGNAL_FUNC(preferences_apply), (void *)handles);
	dw_signal_connect(cancelbutton, "clicked", DW_SIGNAL_FUNC(generic_cancel), (void *)handles);

	dw_window_set_usize(entrywindow, 350, 350);

	dw_window_show(entrywindow);

	switch(fire)
	{
	case 1:
		tmp = 15;
		break;
	case 2:
		tmp = 14;
		break;
	default:
		tmp = 16;
		break;
	}

	dw_checkbox_set(handles[tmp], 1);

    return 0;
}

int _System plugin_deinit(int unload)
{
	if(usedive)
		DiveShutdown();

	if(hpsDive  != NULLHANDLE)
		GpiSelectPalette(hpsDive, NULLHANDLE);
	if(hpalDive != NULLHANDLE)
		GpiDeletePalette(hpalDive);
	if(hpsDive != NULLHANDLE)
		GpiDestroyPS(hpsDive);

	WinDestroyWindow(hwndClient);

	free(last);
	free(bmax);
	free(current);
	free(maxage);
	if(DiveColors)
	{
		free(DiveColors);
		DiveColors = NULL;
	}

	return 0;
}

void CreateMonoColorTable(HPS hps, int r, int g, int b)
{
	short int red = 0, green = 0, blue = 100;
	LONG alTable[17];

	GpiQueryLogColorTable(hps, 0L, 0L, 17L, alTable);

	if(r)
		red = 255;
	else
		red = 0;

	if(g)
		green = 255;
	else
		green = 0;

	if(b)
		blue = 255;
	else
		blue = 0;

	alTable[16] = (red * 65536) + (green * 256) + blue;

	GpiCreateLogColorTable(hps, 0L, LCOLF_CONSECRGB, 0, 17, alTable);
}

void CreateColorTable(HPS hps, int maxcolors, int r, int g, int b)
{
	short int red = 0, green = 0, blue = 100;
	LONG alTable[100];
	/* We want to start 1/4 up the color scale
	 * so we using 192 (256-64) instead of 256.
	 */
	float m = 192/maxcolors;
	int z;

	GpiQueryLogColorTable(hps, 0L, 0L, 100L, alTable);

	for(z=0;z<maxcolors;z++)
	{
		if(r)
			red = (BYTE)((z*m) + 63);
		else
			red = 0;

		if(g)
			green = (BYTE)((z*m) + 63);
		else
			green = 0;

		if(b)
			blue = (BYTE)((z*m) + 63);
		else
			blue = 0;

		alTable[z+16] = (red * 65536) + (green * 256) + blue;
	}

	GpiCreateLogColorTable(hps, 0L, LCOLF_CONSECRGB, 0, 100, alTable);
}

void GpiDrawBlock(HPS hps, int x, int y, int width, int height, int color)
{
	POINTL ptl;
	int newy = (cy - y) - 1;
	int newcolor = color + 14;

	if(color == 0)
		newcolor = CLR_BLACK;
	else if(color == 1)
		newcolor = CLR_WHITE;

	GpiSetColor(hps, newcolor);

	ptl.x = x;
	ptl.y = newy;

	GpiMove(hps, &ptl);

	ptl.x = (x + width) - 1;
    ptl.y = (newy + height) - 1;

	GpiBox(hps, DRO_FILL, &ptl, 0, 0);
}

/*
 * GPI implementation of the analyzer renderer.
 */
void GpiRender(HWND hwnd)
{
	int z, max;
	float scale;
	HPS hps = WinGetPS(hwnd);

	max = specana_dobands(current);

	if(boost)
		scale = (float)max/(cy*boostval);
	else
		scale = (float)max/cy;

	/* We want a gradient */
	if(gradient)
		CreateColorTable(hps, cy, barred, bargreen, barblue);
	else
		CreateMonoColorTable(hps, barred, bargreen, barblue);


#ifdef GPI_FULL_CLEAR
	GpiDrawBlock(hps, 0, cy - 1, cx, cy, 0);
#endif

	for(z=0;z<samples;z++)
	{
		/* double check that we aren't off the screen */
		int lastheight = last[z]/scale;
		int thisheight = current[z]/scale;

		if(lastheight > cy)
			lastheight = cy;

		if(thisheight > cy)
			thisheight = cy;

#ifndef GPI_FULL_CLEAR
		/* Erase any old remnants if the current value
		 * is lower than the last value.
		 */
		if(thisheight < lastheight)
		{
			GpiDrawBlock(hps, z * (blockwidth + spacing), cy - thisheight,
						 blockwidth,  (lastheight - thisheight) + 1,
						 0);
		}
#endif

		/* Draw the bar */
		if(gradient)
		{
			int t, start = cy - thisheight;

			for(t=0;t<thisheight;t++)
			{
				if(fire == 1)
					GpiDrawBlock(hps, z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  (cy-t)+1);
				else if(fire == 2)
					GpiDrawBlock(hps, z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  t+2);
				else
					GpiDrawBlock(hps, z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  (thisheight-t)+1);

			}
		}
		else
		{
			if(thisheight)
			{
				GpiDrawBlock(hps, z * (blockwidth + spacing), cy - 1,
							  blockwidth, thisheight,
							  2);
			}
		}

		if(showpeaks)
		{
			int height;

			/* If we have a new peak save it */
			if(current[z] > bmax[z] || maxage[z] > falloff)
			{
#ifndef GPI_FULL_CLEAR
				/* Erase the old peaks */
				height = bmax[z]/scale;

				/* double check that we aren't off the screen */
				if(height > cy)
					height = cy;

				GpiDrawBlock(hps, z * (blockwidth + spacing), cy - height,
							  blockwidth, 1,
							  0);
#endif

				/* If we are over the falloff start
				 * decreasing the max value exponentially.
				 */
				if(maxage[z] > falloff && current[z] < bmax[z])
				{
					bmax[z] -= (maxage[z]-falloff)*scale;
					maxage[z]++;
				}
				else
				{
					/* Otherwise we reset the current peak to
					 * the new peak, and reset the age.
					 */
					bmax[z] = current[z];
					maxage[z] = 0;
				}
			}
			else
				maxage[z]++;

			height = bmax[z]/scale;

			/* Draw the peaks */
			if(height)
			{
				/* double check that we aren't off the screen */
				if(height > cy)
					height = cy;

				GpiDrawBlock(hps, z * (blockwidth + spacing), cy - height,
							  blockwidth, 1,
							  1);
			}
		}
	}

	WinReleasePS(hps);

	memcpy(last, current, sizeof(float) * samples);
}

void DiveCreateColorTable(int maxcolors, int r, int g, int b)
{
	ULONG palcolors = maxcolors + 2;
	/* We want to start 1/4 up the color scale
	 * so we using 192 (256-64) instead of 256.
	 */
	float m = 192/maxcolors;
	int z;

	DiveColors[0] = rgbBlack;
	DiveColors[1] = rgbWhite;

	for(z=0;z<maxcolors;z++)
	{
		if(r)
			DiveColors[z+2].bRed = (BYTE)((z*m) + 63);
		else
			DiveColors[z+2].bRed = 0;

		if(g)
			DiveColors[z+2].bGreen = (BYTE)((z*m) + 63);
		else
			DiveColors[z+2].bGreen = 0;

		if(b)
			DiveColors[z+2].bBlue = (BYTE)((z*m) + 63);
		else
			DiveColors[z+2].bBlue = 0;
	}
	hpalDive = GpiCreatePalette(hab,
								LCOL_PURECOLOR,
								LCOLF_CONSECRGB,
								palcolors,
								(PULONG)DiveColors);

	GpiSelectPalette(hpsDive, hpalDive);
	WinRealizePalette(hwndClient, hpsDive, &palcolors);
}

void DiveInit(void)
{
	APIRET rc = 0;

	if(DiveOpen(&hDive, FALSE, 0))
	{
		dw_messagebox("Nuclear Visuals", DW_MB_OK | DW_MB_ERROR, "Error initializing DIVE.");
		samples = 0;
		return;
	}

#if 0
	rc = DiveQueryCaps(&DiveCaps, DIVE_BUFFER_SCREEN);

	if(rc || DiveCaps.ulDepth < 8)
	{
		dw_messagebox("Nuclear Visuals", DW_MB_OK | DW_MB_ERROR, "DIVE is present but cannot be used.");
		DiveClose(hDive);
		hDive = NULLHANDLE;
		samples = 0;
		return;
	}
#endif

	DiveAllocImageBuffer(hDive, &DiveBuffer, FOURCC_LUT8,
						 cx, cy, 0, NULL);

	DiveCreateColorTable(cy, barred, bargreen, barblue);

	DiveSetSourcePalette(hDive, 0, cy + 2, (PBYTE)DiveColors);
	DiveSetDestinationPalette(hDive, 0, cy + 2, 0);

	WinSetVisibleRegionNotify(hwndClient, TRUE);
	WinPostMsg(hwndClient, WM_VRNENABLED, 0L, 0L);

	/* This should be 1 bytes per pixel in LUT8 */
	fBufLen = cx * cy;
	fBufScanLineBytes = cx;
	fBufScanLines = cy;
}

void DiveShutdown(void)
{
	WinSetVisibleRegionNotify(hwndClient, FALSE);

	if(DiveBuffer)
		DiveFreeImageBuffer(hDive, DiveBuffer);
	if(hDive != NULLHANDLE)
		DiveClose(hDive);

	DiveBuffer = 0;
	hDive = NULLHANDLE;
}

void DiveDrawBlock(int x, int y, int width, int height, int color)
{
	int b, s, offset;

	for(b=0;b<width;b++)
	{
		for(s=0;s<height;s++)
		{
			offset = (x+b) + ((y + s) * fBufScanLineBytes);

			/* Make sure we aren't going out of our buffer */
			if(offset >= fBufLen)
				return;

			fBuf[offset] = (BYTE)color;
		}
	}
}

/*
 * DIVE implementation of the analyzer renderer.
 */
void DiveRender(HWND hwnd)
{
	int z, max;
	float scale;

	max = specana_dobands(current);

	if(boost)
		scale = (float)max/(cy*boostval);
	else
		scale = (float)max/cy;

	DiveBeginImageBufferAccess(hDive, DiveBuffer, &fBuf, &fBufScanLineBytes, &fBufScanLines);
	fBufLen = fBufScanLineBytes * fBufScanLines;

#ifdef FULL_CLEAR
	memset(fBuf, 0, fBufLen);
#endif

	for(z=0;z<samples;z++)
	{
		/* double check that we aren't off the screen */
		int lastheight = last[z]/scale;
		int thisheight = current[z]/scale;

		if(lastheight > cy)
			lastheight = cy;

		if(thisheight > cy)
			thisheight = cy;

#ifndef FULL_CLEAR
		/* Erase any old remnants if the current value
		 * is lower than the last value.
		 */
		if(thisheight < lastheight)
		{
			DiveDrawBlock(z * (blockwidth + spacing), cy - thisheight,
						  blockwidth,  (lastheight - thisheight) + 1,
						  0);
		}
#endif

		/* Draw the bar */
		if(gradient)
		{
			int t, start = cy - thisheight;

			for(t=0;t<thisheight;t++)
			{
				if(fire == 1)
					DiveDrawBlock(z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  (cy-t)+1);
				else if(fire == 2)
					DiveDrawBlock(z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  t+2);
				else
					DiveDrawBlock(z * (blockwidth + spacing), start + t,
								  blockwidth, 1,
								  (thisheight-t)+1);

			}
		}
		else
		{
			if(thisheight)
			{
				DiveDrawBlock(z * (blockwidth + spacing), cy - thisheight,
							  blockwidth, thisheight,
							  cy);
			}
		}

		if(showpeaks)
		{
			int height;

			/* If we have a new peak save it */
			if(current[z] > bmax[z] || maxage[z] > falloff)
			{
#ifndef FULL_CLEAR
				/* Erase the old peaks */
				height = bmax[z]/scale;

				/* double check that we aren't off the screen */
				if(height > cy)
					height = cy;

				DiveDrawBlock(z * (blockwidth + spacing), cy - height,
							  blockwidth, 1,
							  0);
#endif

				/* If we are over the falloff start
				 * decreasing the max value exponentially.
				 */
				if(maxage[z] > falloff && current[z] < bmax[z])
				{
					bmax[z] -= (maxage[z]-falloff)*scale;
					maxage[z]++;
				}
				else
				{
					/* Otherwise we reset the current peak to
					 * the new peak, and reset the age.
					 */
					bmax[z] = current[z];
					maxage[z] = 0;
				}
			}
			else
				maxage[z]++;

			height = bmax[z]/scale;

			/* Draw the peaks */
			if(height)
			{
				/* double check that we aren't off the screen */
				if(height > cy)
					height = cy;

				DiveDrawBlock(z * (blockwidth + spacing), cy - height,
							  blockwidth, 1,
							  1);
			}
		}
	}

	memcpy(last, current, sizeof(float) * samples);

	DiveEndImageBufferAccess(hDive, DiveBuffer);

	DiveBlitImage(hDive, DiveBuffer, DIVE_BUFFER_SCREEN);

}

MRESULT EXPENTRY PlugWinProc (HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
	switch (msg)
	{

	case DM_DRAGOVER:
	case DM_DROP:
	case 0x041f:
	case 0x041e:
	case WM_CONTEXTMENU:
	case WM_BUTTON2MOTIONSTART:
		WinSendMsg(hwndPM123, msg, mp1, mp2);
		break;

	case WM_TIMER:
		if(procs.decoder_playing())
		{
			if(usedive)
				DiveRender(hwnd);
			else
				GpiRender(hwnd);
		}
		break;
	case WM_REALIZEPALETTE:
		/* This tells DIVE that the physical palette may have changed. */
		if(usedive)
			DiveSetDestinationPalette(hDive, 0, 256, 0);
		break;
	case WM_VRNDISABLED:
		if(usedive)
			DiveSetupBlitter (hDive, 0);
		break;
	case WM_VRNENABLED:
		if(usedive)
		{
			HPS hps = WinGetPS(hwnd);
			HRGN hrgn = GpiCreateRegion(hps, 0L, NULL );
			SWP swp;
			RECTL rcls[50];
			RGNRECT rgnCtl;
			SETUP_BLITTER SetupBlitter;
			POINTL pointl;

			if(hrgn)
			{
				WinQueryVisibleRegion(hwnd, hrgn);
				rgnCtl.ircStart = 0;
				rgnCtl.crc = 50;
				rgnCtl.ulDirection = 1;

				/* Get the all ORed rectangles */
				if(GpiQueryRegionRects(hps, hrgn, NULL, &rgnCtl, rcls))
				{
					/* Now find the window position and size, relative to parent. */
					WinQueryWindowPos(hwndClient, &swp);

					/* Convert the point to offset from desktop lower left. */
					pointl.x = swp.x;
					pointl.y = swp.y;
					WinMapWindowPoints(hwndPM123, HWND_DESKTOP, &pointl, 1);

					/* Tell DIVE about the new settings. */
					SetupBlitter.ulStructLen = sizeof(SETUP_BLITTER);
					SetupBlitter.fccSrcColorFormat = FOURCC_LUT8;
					SetupBlitter.ulSrcWidth = cx;
					SetupBlitter.ulSrcHeight = cy;
					SetupBlitter.ulSrcPosX = 0;
					SetupBlitter.ulSrcPosY = 0;
					SetupBlitter.fInvert = FALSE;
					SetupBlitter.ulDitherType = 1;

					SetupBlitter.fccDstColorFormat = FOURCC_SCRN;
					SetupBlitter.ulDstWidth = swp.cx;
					SetupBlitter.ulDstHeight = swp.cy;
					SetupBlitter.lDstPosX = 0;
					SetupBlitter.lDstPosY = 0;
					SetupBlitter.lScreenPosX = pointl.x;
					SetupBlitter.lScreenPosY = pointl.y;
					SetupBlitter.ulNumDstRects = rgnCtl.crcReturned;
					SetupBlitter.pVisDstRects = rcls;
					DiveSetupBlitter(hDive, &SetupBlitter);
				}
				else
					DiveSetupBlitter(hDive, 0);

				GpiDestroyRegion(hps, hrgn);
			}
			WinReleasePS(hps);
		}
		break;
	}
	return WinDefWindowProc (hwnd, msg, mp1, mp2);
}

HWND _System vis_init(PVISPLUGININIT initdata)
{
	int z = 1, wantedsamples;
	SIZEL sizl;

	bands = 0;

	WinRegisterClass(initdata->hab,
					 "NuclearPlugin",
					 PlugWinProc,
					 CS_SIZEREDRAW, 0);

	hwndClient = WinCreateWindow(initdata->hwnd,
								 "NuclearPlugin",
								 "PM123 Nuclear Visual Plug-in",
								 WS_VISIBLE,
								 initdata->x,
								 initdata->y,
								 initdata->cx,
								 initdata->cy,
								 initdata->hwnd,
								 HWND_TOP,
								 initdata->id,
								 NULL,
								 NULL);


	/* Save the PM123 window handle so we can bounce
	 * messages to it when we don't want to deal with
	 * them.
	 */
	hwndPM123 = initdata->hwnd;

	/* Save copies of the function pointers for our use */
	memcpy(&procs, initdata->procs, sizeof(PLUGIN_PROCS));
	specana_init = initdata->procs->specana_init;
	specana_dobands = initdata->procs->specana_dobands;

	cx = initdata->cx;
	cy = initdata->cy;
	hab = initdata->hab;

	loadconfig();

	samples = (int)(cx / (blockwidth+spacing));
	wantedsamples = (int)((cx / (blockwidth+spacing))*(100/percentage));

	/* Find the ideal number of bars (bands) for our analyzer */
	while(bands < wantedsamples)
	{
		bands = specana_init(z);
		z++;
		if(z < 1 || z > 4097)
		{
			dw_messagebox("Nuclear Visuals", DW_MB_OK | DW_MB_ERROR, "Error getting spectrum analyzer");
			samples = 0;
			break;
		}
	}

	maxage = (int *)calloc(bands, sizeof(int));
	bmax = (float *)calloc(bands, sizeof(float));
	last = (float *)calloc(bands, sizeof(float));
	current = (float *)calloc(bands, sizeof(float));
	if(usedive)
	{
		DiveColors = (RGB2 *)calloc(cy+2, sizeof(RGB2));
		DiveInit();
		WinSetVisibleRegionNotify(hwndClient, TRUE);
		WinPostMsg(hwndClient, WM_VRNENABLED, 0L, 0L);
	}

	hdcClient = WinOpenWindowDC(hwndClient);
	hpsDive = GpiCreatePS(hab, hdcClient, &sizl,
						  PU_PELS | GPIF_DEFAULT | GPIT_MICRO | GPIA_ASSOC);

	dw_init(FALSE, 0, NULL);

	clear_analyzer();

	WinStartTimer(hab, hwndClient, 449, updatefreq);

	return(hwndClient);
}

